# Plugin for TWiki Enterprise Collaboration Platform, http://TWiki.org/
#
# Copyright (C) 2013-2015 Wave Systems Corp.
# Copyright (C) 2013-2015 Peter Thoeny & TWiki Contributors
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version. For
# more details read LICENSE in the root of this distribution.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details, published at
# http://www.gnu.org/copyleft/gpl.html
#
# As per the GPL, removal of this notice is prohibited.

package TWiki::Plugins::NotifyAuthorsPlugin::Core;

# =========================
sub new {
    my ( $class, $debug ) = @_;

    my $this = {
          Debug          => $debug,
          ButtonText     => _getPref( 'ButtonText', 'Notify authors' ),
          TooltipText    => _getPref( 'TooltipText',
            'Notify via e-mail all authors who contributed to this page' ),
          HelpText       => _getPref( 'HelpText',
            'Notify %NOTIFYAUTHORLIST% by e-mail to solicit feedback on this topic.' ),
          NotifiedText   => _getPref( 'NotifiedText',
            '%NOTIFYAUTHORLIST% have been notified by e-mail.' ),
          AnchorName     => _getPref( 'AnchorName', '' ),
          AuthorPatterns => _getPref( 'AuthorPatterns',
            '\\-\\- +%USERSWEB%\\.(\\w+), \\%USERSIG{[ "]*(\\w+), ' .
            '\\%BUBBLESIG{[ "]*(\\w+),' ),
          SummaryFormat  => _getPref( 'SummaryFormat', '$summary(100)' ),
          ButtonTemplate => _getPref( 'ButtonTemplate', 'notifyauthorsbutton' ),
          EmailTemplate  => _getPref( 'EmailTemplate', 'notifyauthorsemail' ),
          EmailText      => '',
          UseEmailField  => _getPref( 'UseEmailField', 0 ),
          LogAction      => _getPref( 'LogAction', 0 ),
        };
    bless( $this, $class );

    #$this->{Debug} = 1; # Uncomment to enable debugging just in this file

    TWiki::Func::writeDebug( "- NotifyAuthorsPlugin Core constructor" ) if $this->{Debug};

    return $this;
}

# =========================
sub VarNOTIFYAUTHORS
{
    my ( $this, $session, $params, $topic, $web ) = @_;

    my $query = TWiki::Func::getCgiQuery();
    my $urlParamAction = '';
    my $authorList = '';
    my $emailBody = '';
    if( $query ) {
        $urlParamAction = $query->param('notifyauthors_action')    || '';
        $authorList     = $query->param('notifyauthors_authors')   || '';
        $emailBody      = $query->param('notifyauthors_emailbody') || '';
    }
    my $action = $params->{action} || $urlParamAction || 'showbutton';
    TWiki::Func::writeDebug( "- NotifyAuthorsPlugin NOTIFYAUTHORS, action=$action" )
      if $this->{Debug};
    $this->{anchor} = $params->{anchor} || $this->{AnchorName};
    $this->{anchor} = '#' . $this->{anchor} if( $this->{anchor} );
    $this->{text}   = '<a name="NotifyAuthorsButton"></a>';
    $this->{help}   = '';
    $this->{note}   = '';
    $this->{error}  = '';

    if( $action eq 'showbutton' ) {
        $this->_actionShowButton( $web, $topic );
    } elsif( $action eq 'getemailtemplate' ) {
        $this->_actionGetEmailTemplate( $web, $topic, $params->{encode}, 0 );
    } elsif( $action eq 'getemailsubject' ) {
        $this->_actionGetEmailTemplate( $web, $topic, $params->{encode}, 1 );
    } elsif( $action eq 'notify' ) {
        $this->_actionNotify( $web, $topic, $authorList, $emailBody );
    }

    my $text = $this->{text};
    my $style = 'margin: 0 0.5em 0 0; padding: 0.3em 0.5em;';
    if( $this->{help} ) {
        $text .= "<span style='color: #666666; $style'> "
               . $this->{help} . " </span>"
    }
    if( $this->{note} ) {
        $text .= "<span style='background-color: #feffc2; $style'> "
               . $this->{note}
               . " [[$web.$topic" . $this->{anchor} . "][OK]] </span>";
    }
    if( $this->{error} ) {
        $text .= "<span style='background-color: #ff9594; $style'> *Error:* "
               . $this->{error}
               . " [[$web.$topic" . $this->{anchor} . "][OK]] </span>"
    }
    return $text;
}

# =========================
sub _actionShowButton
{
    my ( $this, $web, $topic ) = @_;
    my @authors = $this->_getAuthorsFromTopic( $web, $topic );
    my $notifyAuthors = '<nop>' . join( ', <nop>', @authors );

    my $text = TWiki::Func::readTemplate( $this->{ButtonTemplate} );
    $text =~ s/%NOTIFYAUTHORLIST%/$notifyAuthors/go;
    $text =~ s/%NOTIFYANCHOR%/$this->{anchor}/go;
    $text =~ s/%NOTIFYBUTTONTEXT%/$this->{ButtonText}/go;
    $text =~ s/%NOTIFYTOOLTIP%/$this->{TooltipText}/go;
    $this->{text} .= $text;

    $this->{help} = $this->{HelpText};
    $this->{help} =~ s/%NOTIFYAUTHORLIST%/$notifyAuthors/go;
    return;
}

# =========================
sub _actionGetEmailTemplate
{
    my ( $this, $web, $topic, $encode, $getSubject ) = @_;

    my $text = $this->{EmailText} ||
             ( $this->{EmailText} = TWiki::Func::readTemplate( $this->{EmailTemplate} ) );
    if( $getSubject ) {
        $text =~ s/^.*?Subject: *([^\n]*).*$/$1/os;
    } else {
        # cut e-mail header from template
        $text =~ s/^.*?\n\n//os;
    }
    $text = $this->_expandVariables( $web, $topic, $text );
    if( $encode ) {
        if( $TWiki::Plugins::VERSION >= 6.0 ) {
            $text = TWiki::Func::entityEncode( $text, " \n\r" );
        } else {
            $text = TWiki::entityEncode( $text, " \n\r" );
        }
    }
    $this->{text} = $text;
    return;
}

# =========================
sub _actionNotify
{
    my ( $this, $web, $topic, $authorList, $emailBody ) = @_;

    # Identify authors and gather e-mail addresses
    my @authors = split( /, */, $authorList );
    @$authors = $this->_getAuthorsFromTopic( $web, $topic ) unless scalar( @authors );
    my $toEmails = '';
    if( $this->{UseEmailField} ) {
        # Get e-mails from Email form field of user profile topics
        my $result = TWiki::Func::searchInWebContent(
          '%META:FIELD{name="Email"',
          $TWiki::cfg{UsersWebName},
          \@authors, {} );
        my @emails = ();
        foreach my $user (keys %$result ) {
            foreach my $line ( @{$result->{$user}} ) {
                if( $line =~ s/^.*value="([^"]*)".*$/$1/o ) {
                    push( @emails, "$user <$line>" );
                }
            }
        }
        $toEmails = join( ', ', @emails );

    } else {
        # Get e-mails from password system
        my @emails = ();
        foreach my $user ( @authors ) {
           foreach my $mail ( TWiki::Func::wikinameToEmails( $user ) ) {
               push( @emails, "$user <$mail>" );
           }           
        }
        $toEmails = join( ', ', @emails );
    }

    unless( scalar @authors ) {
        $this->{note} = "No authors could be found in !$topic topic.";
        return;
    }

    my $text = $this->{EmailText} || 
             ( $this->{EmailText} = TWiki::Func::readTemplate( $this->{EmailTemplate} ) );
    if( $emailBody ) {
        # cut e-mail body from template and use body from parameter
        $text =~ s/\n\n.*$//os;
        $text .= "\n\n$emailBody";
    }
    $text = $this->_expandVariables( $web, $topic, $text, $toEmails );

    $this->_sendEmail( $web, $topic, $text );
    unless( $this->{error} ) {
        my $notifyAuthors = '<nop>' . join( ', <nop>', @authors );
        $this->{note} .= $this->{NotifiedText};
        $this->{note} =~ s/%NOTIFYAUTHORLIST%/$notifyAuthors/go;
        if( $this->{LogAction} && $TWiki::Plugins::VERSION >= 1.4 ) {
            my $extra = 'notified: ' . join( ', ', @authors );
            TWiki::Func::writeLog( 'notifyauthors', $extra, $web, $topic );
        }
    }

    return;
}

# =========================
sub _getAuthorsFromTopic {
    my( $this, $web, $topic ) = @_;

    my $regex = $this->{AuthorPatterns};
    $regex =~ s/\%(USERSWEB|MAINWEB)\%/$TWiki::cfg{UsersWebName}/g;
    my @regexes = split( /, */, $regex );
    $regex =~ s/, */|/g;

    my %seen;
    my @authors =
      grep { ! $this->{UseEmailField}
            || TWiki::Func::topicExists( $TWiki::cfg{UsersWebName}, $_ ) }
      grep { !$seen{$_}++ }
      sort
      map {
          my @a;
          foreach my $r ( @regexes ) {
              s/^.*?$r.*$/$1/;                       # Keep only name
              s/(Main|\%USERSWEB\%|\%MAINWEB\%)\.//; # cut web prefix if any
              @a = split( /, */, $_ );               # Assume single name or comma-space list
          }
          @a;
      }
      grep { /$regex/ }
      split( /\r?\n/, TWiki::Func::readTopicText( $web, $topic, undef, 1 ) );

    return @authors;
}

# =========================
sub _expandVariables {
    my( $this, $web, $topic, $text, $toEmails ) = @_;

    # For title, use TOPICTITLE if available, else space out topic name
    my $rawTitle = "%TOPICTITLE{ topic=\"$web.$topic\" }%";
    my $title = TWiki::Func::expandCommonVariables( $rawTitle, $topic, $web );
    if( $title eq $rawTitle || $title eq $topic ) {
        $title = "%SPACEOUT{$topic}%";
    }
    my $wikiName = TWiki::Func::getWikiName();

    $text =~ s/%NOTIFYSUMMARY%/$this->_getTopicSummary( $web, $topic )/geo;
    $text =~ s/%NOTIFYFROM%/$wikiName/go;
    $text =~ s/%NOTIFYADDRESSES%/$toEmails/go if( $toEmails );
    $text =~ s/%NOTIFYWEB%/$web/go;
    $text =~ s/%NOTIFYTOPIC%/$topic/go;
    $text =~ s/%NOTIFYTITLE%/$title/go;
    $text = TWiki::Func::expandCommonVariables( $text, $topic, $web );
    return $text;
}

# =========================
sub _sendEmail {
    my( $this, $web, $topic, $text ) = @_;

    $this->{error} = TWiki::Func::sendEmail( $text );
    $this->{note} .= "<verbatim>$text</verbatim>" if( $this->{Debug} );
    return;
}

# =========================
sub _getTopicSummary
{
    my ( $this, $web, $topic ) = @_;
    my $var = "\%SEARCH{ \"$topic\" scope=\"topic\" web=\"$web\" topic=\"$topic\""
            . ' nonoise="on" format="' . $this->{SummaryFormat} . '" }%';
    my $text = TWiki::Func::expandCommonVariables( $var, $topic, $web );
    $text =~ s/<nop>//go; # Remove <nop> tags
    return $text;
}

# =========================
sub _getPref
{
    my ( $name, $default ) = @_;
    my $value = TWiki::Func::getPreferencesValue( uc( "NotifyAuthorsPlugin_$name" ) )
             || $TWiki::cfg{Plugins}{NotifyAuthorsPlugin}{$name}
             || $default;
    return $value;
}

# =========================
1;
